"""Initialize pkg_support_resistance layer modules"""

from .vanilla.vanilla_algo import VanillaSupportResistance  # noqa: F401
from .kmeans.kmeans_algo import KMeansSupportResistance  # noqa: F401
from .dbscan.dbscan_algo import DBSCANSupportResistance  # noqa: F401

# Exports Data-Set
from .data_set.data_extraction import sr_input_example, open_json_file

__version__ = "0.4.0"



"""Data-extraction module"""

from typing import Dict, Any, List
import json


def open_json_file(input_file_path: str, mode: str = "r") -> List[Any] | Dict[str, Any]:
    """Open json file"""

    with open(file=input_file_path, mode=mode) as file:
        data = json.load(file)
    return data


# Call function to extract data from JSON file
sr_input_example: Dict[str, List[Any]] = open_json_file(
    input_file_path="src/pkg_support_resistance/data_set/example.json"
)



"""DBSCAN ML algorithms test module"""

from src.pkg_support_resistance import DBSCANSupportResistance
from src.pkg_support_resistance.data_set import sr_input_example
from .expected_output_test import dbscan_sr_exec_pipeline_expected_output_test


def test_dbscan_algorithms_exec_pipeline():
    """
    Test dbscan algorithms the calculation of support/resistance process.
    This test verifies that the function exec_pipeline,
    """

    result: list[dict] = DBSCANSupportResistance.exec_pipeline(
        input_data=sr_input_example, eps=1000, min_samples=1
    )
    assert result == dbscan_sr_exec_pipeline_expected_output_test



"""KMeans ML algorithms module"""

from typing import Any, Dict, List, Union

import pandas as pd
from sklearn.cluster import KMeans


class KMeansSupportResistance:
    """KMeans Clustering Support Resistance (unsupervised algorithms) class"""

    @staticmethod
    def is_cythonized():
        """
        Check if the module is cytotonized?
        The cytonized model does not represent a significant performance improvement,
        because sklear is already optimized.
        """

        return False

    @classmethod
    def __preprocess_data(
        cls, input_data: Dict[str, List[Union[float, int]]]
    ) -> pd.DataFrame:
        """
        Preprocess data.
        Return: pd.DataFrame.
        I.e.:
            open        close      high       low     volume    avg_price
        0    42780.00  42834.94  42901.10  42730.03  173.63661  42811.5175
        1    42834.94  42961.84  42986.06  42795.41  169.47648  42894.5625
        2    42961.83  43069.98  43160.86  42914.13  298.77136  43026.7000
        3    43070.30  43139.40  43159.74  42988.00  221.97150  43089.3600
        4    43139.40  43303.82  43333.00  43115.00  398.10382  43222.8050
        ...
        """

        df: pd.DataFrame = pd.DataFrame(data=input_data)
        df["avg_price"] = df[["open", "close", "high", "low"]].mean(axis=1)
        return df

    @classmethod
    def __apply_kmeans(
        cls, df: pd.DataFrame, n_clusters: int | None = 5
    ) -> pd.DataFrame:
        """
        Apply kmeans.
        Return: pd.DataFrame.
        I.e.:
            open         close      high       low     volume   avg_price   cluster
        0    42780.00  42834.94  42901.10  42730.03  173.63661  42811.5175        4
        1    42834.94  42961.84  42986.06  42795.41  169.47648  42894.5625        4
        2    42961.83  43069.98  43160.86  42914.13  298.77136  43026.7000        4
        3    43070.30  43139.40  43159.74  42988.00  221.97150  43089.3600        4
        4    43139.40  43303.82  43333.00  43115.00  398.10382  43222.8050        4
        ...
        """

        kmeans = KMeans(n_clusters=n_clusters, random_state=0).fit(df[["avg_price"]])
        df["cluster"] = kmeans.labels_
        return df

    @classmethod
    def __calculate_cluster(
        cls, df: pd.DataFrame, n_clusters: int
    ) -> List[Dict[str, Any]]:
        """
        Calculate cluster.
        Return: List[Dict[str, Any]].
        I.e.: [
            {'pivotPrice': 41247.07, 'limitsDown': 40965.2125, 'limitsUp': 41334.82750000001, 'accumulatedVolume': 11484.42265, 'score': 25},
            {'pivotPrice': 41541.4925, 'limitsDown': 41350.675, 'limitsUp': 41683.4625, 'accumulatedVolume': 16764.66682, 'score': 44},
            {'pivotPrice': 41740.695, 'limitsDown': 41685.535, 'limitsUp': 42090.5675, 'accumulatedVolume': 25677.978769999998, 'score': 54},
            {'pivotPrice': 42616.125, 'limitsDown': 42105.2425, 'limitsUp': 42692.68749999999, 'accumulatedVolume': 28562.038430000004, 'score': 68},
            {'pivotPrice': 42939.847499999996, 'limitsDown': 42811.5175, 'limitsUp': 43384.869999999995, 'accumulatedVolume': 29510.197329999995, 'score': 65},
            {'pivotPrice': 43669.6875, 'limitsDown': 43419.9975, 'limitsUp': 44002.247500000005, 'accumulatedVolume': 25561.89504, 'score': 64},
            {'pivotPrice': 45788.325, 'limitsDown': 44330.0825, 'limitsUp': 45788.325, 'accumulatedVolume': 12449.31098, 'score': 13},
            {'pivotPrice': 46947.585, 'limitsDown': 46947.585, 'limitsUp': 47633.6625, 'accumulatedVolume': 40150.74441, 'score': 55},
            {'pivotPrice': 47886.935000000005, 'limitsDown': 47651.009999999995, 'limitsUp': 48355.61749999999, 'accumulatedVolume': 39565.819639999994, 'score': 111}
        ]
        """

        cluster_info = []
        for i in range(n_clusters):
            cluster_data = df[df["cluster"] == i]
            limitsDown = cluster_data["avg_price"].min()
            limitsUp = cluster_data["avg_price"].max()
            accumulatedVolume = cluster_data["volume"].sum()

            max_volume_idx = cluster_data["volume"].idxmax()
            pivotPrice = cluster_data.loc[max_volume_idx, "avg_price"]
            score = len(cluster_data)

            cluster_info.append(
                {
                    "cluster": i,
                    "pivotPrice": pivotPrice,
                    "limitsDown": limitsDown,
                    "limitsUp": limitsUp,
                    "score": score,
                    "accumulatedVolume": accumulatedVolume,
                }
            )

        cluster_sorted: list[dict] = sorted(cluster_info, key=lambda x: x["pivotPrice"])
        cluster_result: list[dict] = [
            {k: v for k, v in d.items() if k != "cluster"} for d in cluster_sorted
        ]
        return cluster_result

    @classmethod
    def exec_pipeline(
        cls, input_data: Dict[str, List[Union[float, int]]], n_clusters: int | None = 5
    ) -> List[Dict[str, Any]]:
        """
        Calculate support/resistance process pipeline.
        Return: list[dict[str, Union[int, float]]]. Support/Resistance calculated data. I.e.: [
            {'pivotPrice': 41247.07, 'limitsDown': 40965.2125, 'limitsUp': 41334.82750000001, 'accumulatedVolume': 11484.42265},
            {'pivotPrice': 41541.4925, 'limitsDown': 41350.675, 'limitsUp': 41683.4625, 'accumulatedVolume': 16764.66682},
            {'pivotPrice': 41740.695, 'limitsDown': 41685.535, 'limitsUp': 42090.5675, 'accumulatedVolume': 25677.978769999998},
            {'pivotPrice': 42616.125, 'limitsDown': 42105.2425, 'limitsUp': 42692.68749999999, 'accumulatedVolume': 28562.038430000004},
            {'pivotPrice': 42939.847499999996, 'limitsDown': 42811.5175, 'limitsUp': 43384.869999999995, 'accumulatedVolume': 29510.197329999995},
            {'pivotPrice': 43669.6875, 'limitsDown': 43419.9975, 'limitsUp': 44002.247500000005, 'accumulatedVolume': 25561.89504},
            {'pivotPrice': 45788.325, 'limitsDown': 44330.0825, 'limitsUp': 45788.325, 'accumulatedVolume': 12449.31098},
            {'pivotPrice': 46947.585, 'limitsDown': 46947.585, 'limitsUp': 47633.6625, 'accumulatedVolume': 40150.74441},
            {'pivotPrice': 47886.935000000005, 'limitsDown': 47651.009999999995, 'limitsUp': 48355.61749999999, 'accumulatedVolume': 39565.819639999994}
        ]
        """

        df: pd.DataFrame = cls.__preprocess_data(input_data=input_data)
        df: pd.DataFrame = cls.__apply_kmeans(df=df, n_clusters=n_clusters)
        cluster_model: list[dict] = cls.__calculate_cluster(
            df=df, n_clusters=n_clusters
        )
        return cluster_model



"""DBScan ML algorithms module"""

from typing import Any, Dict, List, Union

from sklearn.cluster import DBSCAN
import numpy as np


class DBSCANSupportResistance:
    """DBSCAN Clustering Support Resistance (unsupervised algorithms) class"""

    @staticmethod
    def is_cythonized():
        """
        Check if the module is cytotonized?
        The cytonized model does not represent a significant performance improvement,
        because sklear is already optimized.
        """

        return False

    @classmethod
    def __preprocess_data(
        cls, input_data: Dict[str, List[Union[float, int]]]
    ) -> np.ndarray:
        """
        Preprocess data.
        Return: np.ndarray.
        I.e.: [
            [42780.      42834.94    42901.1     42730.03      173.63661]
            [42834.94    42961.84    42986.06    42795.41      169.47648]
            [42961.83    43069.98    43160.86    42914.13      298.77136]
            ...
            [47566.11    47593.24    47683.7     47544.4       160.1838 ]
            [47593.24    47714.91    47720.      47593.24      129.70661]
            [47714.92    47898.66    47963.      47634.72      391.51242]
        ]
        """

        x: np.ndarray = np.column_stack(
            tup=(
                input_data["open"],
                input_data["close"],
                input_data["high"],
                input_data["low"],
                input_data["volume"],
            )
        )
        return x

    @classmethod
    def __apply_dbscan(
        cls, x: np.ndarray, eps: float = 1000, min_samples: int = 1
    ) -> np.ndarray:
        """
        Apply dbscan.
        Params eps: float. Defines the maximum distance between two samples so that they are considered to be in the same neighborhood.
        Params min_samples: int. Specifies the minimum number of samples needed in a neighborhood for a point to be considered a center point.
        Return: np.ndarray.
        I.e.: [
            0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
            ...
            6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
            6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
        ]
        """

        dbscan = DBSCAN(eps=eps, min_samples=min_samples)
        labels: np.ndarray = dbscan.fit_predict(X=x)
        return labels

    @classmethod
    def __calculate_cluster(
        cls, labels: np.ndarray, x: np.ndarray
    ) -> List[Dict[str, Any]]:
        """
        Calculate cluster.
        Return: List[Dict[str, Any]].
        I.e.: [
            {'pivotPrice': 42476.84449999999, 'limitsDown': 40841.94, 'limitsUp': 43995.1, 'score': 320, 'accumulatedVolume': 137561.19903999998},
            {'pivotPrice': 44007.53, 'limitsDown': 44007.53, 'limitsUp': 44007.53, 'score': 1, 'accumulatedVolume': 2203.5378},
            {'pivotPrice': 44833.46181818182, 'limitsDown': 44623.46, 'limitsUp': 44945.81, 'score': 11, 'accumulatedVolume': 7232.53705},
            {'pivotPrice': 44855.63, 'limitsDown': 44855.63, 'limitsUp': 44855.63, 'score': 1, 'accumulatedVolume': 3013.23613},
            {'pivotPrice': 46449.22, 'limitsDown': 46449.22, 'limitsUp': 46449.22, 'score': 1, 'accumulatedVolume': 6171.46155},
            {'pivotPrice': 47332.14, 'limitsDown': 47332.14, 'limitsUp': 47332.14, 'score': 1, 'accumulatedVolume': 2715.50006},
            {'pivotPrice': 47731.84621951219, 'limitsDown': 46783.89, 'limitsUp': 48353.34, 'score': 164, 'accumulatedVolume': 70829.60243999999}
        ]
        """

        labels_set = set(labels)
        s_r_calculated_list = []
        for label in labels_set:
            if label == -1:
                continue

            cluster = x[labels == label]
            pivot_price = np.mean(cluster[:, 0])
            limits_down = np.min(cluster[:, 0])
            limits_up = np.max(cluster[:, 0])
            score = len(cluster)
            accumulated_volume = np.sum(cluster[:, 4])

            s_r_calculated_list.append(
                {
                    "pivotPrice": pivot_price,
                    "limitsDown": limits_down,
                    "limitsUp": limits_up,
                    "score": score,
                    "accumulatedVolume": accumulated_volume,
                }
            )

        return s_r_calculated_list

    @classmethod
    def exec_pipeline(
        cls,
        input_data: Dict[str, List[Union[float, int]]],
        eps: float = 1000,
        min_samples: int = 1,
    ) -> List[Dict[str, Any]]:
        """
        Calculate support/resistance process pipeline.
        Params eps: float. Defines the maximum distance between two samples so that they are considered to be in the same neighborhood.
        Params min_samples: int. Specifies the minimum number of samples needed in a neighborhood for a point to be considered a center point.

        Return: list[dict[str, Union[int, float]]]. Support/Resistance calculated data. I.e.: [
            {'pivotPrice': 42476.84449999999, 'limitsDown': 40841.94, 'limitsUp': 43995.1, 'score': 320, 'accumulatedVolume': 137561.19903999998},
            {'pivotPrice': 44007.53, 'limitsDown': 44007.53, 'limitsUp': 44007.53, 'score': 1, 'accumulatedVolume': 2203.5378},
            {'pivotPrice': 44833.46181818182, 'limitsDown': 44623.46, 'limitsUp': 44945.81, 'score': 11, 'accumulatedVolume': 7232.53705},
            {'pivotPrice': 44855.63, 'limitsDown': 44855.63, 'limitsUp': 44855.63, 'score': 1, 'accumulatedVolume': 3013.23613},
            {'pivotPrice': 46449.22, 'limitsDown': 46449.22, 'limitsUp': 46449.22, 'score': 1, 'accumulatedVolume': 6171.46155},
            {'pivotPrice': 47332.14, 'limitsDown': 47332.14, 'limitsUp': 47332.14, 'score': 1, 'accumulatedVolume': 2715.50006},
            {'pivotPrice': 47731.84621951219, 'limitsDown': 46783.89, 'limitsUp': 48353.34, 'score': 164, 'accumulatedVolume': 70829.60243999999}
        ]
        """

        x: np.ndarray = cls.__preprocess_data(input_data=input_data)
        labels: np.ndarray = cls.__apply_dbscan(x=x, eps=eps, min_samples=min_samples)
        cluster_model: list[dict] = cls.__calculate_cluster(labels=labels, x=x)
        return cluster_model



"""Vanilla algorithms test module"""

from src.pkg_support_resistance import VanillaSupportResistance
from src.pkg_support_resistance.data_set import sr_input_example
from .expected_output_test import vanilla_sr_exec_pipeline_expected_output_test


def test_vanilla_algorithms_exec_pipeline():
    """
    Test vanilla algorithms the calculation of support/resistance process.
    This test verifies that the function exec_pipeline,
    """

    result: list[dict] = VanillaSupportResistance.exec_pipeline(
        input_data=sr_input_example, cluster_threshold=1
    )
    assert result == vanilla_sr_exec_pipeline_expected_output_test



"""Vanilla algorithms module"""

from typing import Dict, List, Union

import numpy as np


class VanillaSupportResistance:
    """Vanilla Clustering Support Resistance class"""

    @staticmethod
    def is_cythonized():
        """
        Check if the module is cytotonized?
        The cythonized model represents a performance improvement of %13.05.
        """

        return False

    @classmethod
    def __calc_points(cls, data: np.ndarray, extreme: str) -> np.ndarray[np.ndarray]:
        """
        Calculate points (max o min).
        Return: numpy.ndarray. Dtype = [('idx', 'i4'), ('open', 'f8'), ('close', 'f8'), ('high', 'f8'), ('low', 'f8'), ('volume', 'f8')]
        I.e.: [[4.93000000e+02 4.77248100e+04 4.75912400e+04 4.77299900e+04 4.75580000e+04 1.92358790e+02], ...]
        """

        points = []
        extreme_value = data[:, 2] if extreme == "max" else data[:, 3]

        center = extreme_value[1:-1]
        left = extreme_value[:-2]
        right = extreme_value[2:]

        max_condition = (center > left) & (center > right)
        min_condition = (center < left) & (center < right)

        np.where(max_condition)[0] + 1  # Index adjustment
        np.where(min_condition)[0] + 1  # Index adjustment

        idx = None
        if extreme == "max":
            idx = np.where(max_condition)[0] + 1
        elif extreme == "min":
            idx = np.where(min_condition)[0] + 1

        points = np.empty((len(idx), 6))
        for i, idx in enumerate(idx):
            points[i] = [
                idx,
                data[idx, 0],
                data[idx, 1],
                data[idx, 2],
                data[idx, 3],
                data[idx, 4],
            ]

        return points

    @classmethod
    def __clustering(  # pylint: disable=too-many-locals, too-many-branches
        cls, data: np.ndarray, threshold: float, direction: str
    ) -> List[np.ndarray[np.ndarray]]:
        """
        Clustering.
        Return: numpy.ndarray[numpy.ndarray[numpy.ndarray]].
        dtype: [('idx', 'i4'), ('open', 'f8'), ('close', 'f8'), ('high', 'f8'), ('low', 'f8'), ('volume', 'f8')]
        I.e.: [[[493., 47724.81, 47591.24, 47729.99, 47558., 192.35879], ...], ...]
        """

        clusters = []
        high_index = 3  # Index of the 'high' column in the data array
        low_index = 4  # Index of the 'low' column in the data array

        for curr_first in data:  # pylint: disable=too-many-nested-blocks
            cluster = []
            if direction == "resistance":
                h0_first = curr_first[high_index]
            elif direction == "support":
                h0_first = curr_first[low_index]
            else:
                h0_first = None

            cluster.append(curr_first)
            for curr_second in data:
                if direction == "resistance":
                    h0_second = curr_second[high_index]
                elif direction == "support":
                    h0_second = curr_second[low_index]
                else:
                    h0_second = None

                dif = abs(((h0_first - h0_second) / h0_first) * 100)

                if dif <= threshold:
                    cluster.append(curr_second)

            if cluster:
                if not clusters:
                    clusters.append(np.array(cluster))
                else:
                    push = True
                    clust = (
                        cluster[0][high_index]
                        if direction == "resistance"
                        else cluster[0][low_index]
                    )

                    for curr in clusters:
                        curr_clust = (
                            curr[0][high_index]
                            if direction == "resistance"
                            else curr[0][low_index]
                        )

                        dif = abs(((clust - curr_clust) / clust) * 100)

                        if dif <= threshold:
                            push = False
                            if len(curr) > len(cluster):
                                cluster = curr

                    if push:
                        clusters.append(np.array(cluster))

        return clusters

    @classmethod
    def __recovery_line(  # pylint: disable=too-many-locals
        cls, clusters: List[np.ndarray[np.ndarray]], direction: str
    ) -> np.ndarray:
        """
        Recovery line.
        Return: numpy.ndarray. dtype: [
            ('idx', 'i4'), ('open', 'f8'), ('close', 'f8'), ('high', 'f8'), ('low', 'f8'), ('volume', 'f8')
        ]
        I.e.: [
            (47729.99, 48495.  , 47716.9 , 32, 16145.34576)
            (45000.  , 45000.  , 44990.  ,  4,  3689.66919)
            (43996.5 , 44141.37, 43307.85, 25, 12609.72691)
            (43473.45, 43777.  , 43127.97, 20, 10403.64128)
            (42259.58, 42787.38, 42077.62, 21,  9709.83536)
            (41465.  , 42593.69, 41450.  , 28, 13145.51234)
        ]
        """

        dtype = np.dtype(
            [
                ("pivotPrice", "f8"),
                ("limitsUp", "f8"),
                ("limitsDown", "f8"),
                ("score", "i4"),
                ("accumulatedVolume", "f8"),
            ]
        )

        lines = []
        for curr_cluster in clusters:
            pivot_price = None
            approach = 0
            limits_up = 0
            limits_down = np.inf
            accumulated_volume = 0
            curr_data_price_line = None
            data_line = None

            for idx in range(len(curr_cluster)):
                if direction == "resistance":
                    curr_data_price_line = curr_cluster[idx, 3]
                    data_line = (
                        curr_cluster[idx + 1, 3]
                        if idx + 1 < len(curr_cluster)
                        else None
                    )
                elif direction == "support":
                    curr_data_price_line = curr_cluster[idx, 4]
                    data_line = (
                        curr_cluster[idx + 1, 4]
                        if idx + 1 < len(curr_cluster)
                        else None
                    )

                limits_up = max(limits_up, curr_data_price_line)
                limits_down = min(limits_down, curr_data_price_line)

                if curr_data_price_line >= (data_line if data_line is not None else 0):
                    pivot_price = curr_data_price_line
                else:
                    pivot_price = data_line

                approach += 1
                accumulated_volume += curr_cluster[idx, 5]

            line = (
                pivot_price,
                max(limits_up, pivot_price),
                min(limits_down, pivot_price),
                approach,
                accumulated_volume,
            )

            if len(lines) > 0:
                lines_np = np.array(lines, dtype=dtype)
                existing_line_idx = np.where(lines_np["pivotPrice"] == pivot_price)[0]
            else:
                existing_line_idx = np.array([])

            if len(existing_line_idx) == 0:
                lines.append(line)
            else:
                idx_update = existing_line_idx[0]
                if lines[idx_update][3] < approach:
                    lines[idx_update] = line

        lines_np = np.array(lines, dtype=dtype)
        return np.sort(lines_np, order="pivotPrice")[::-1]

    @classmethod
    def __unify(
        cls,
        src_r: np.ndarray,
        src_s: np.ndarray,
        threshold: float,
        total_lines: np.ndarray | None = np.ndarray,
    ) -> np.ndarray:
        """
        Unify.
        Return: numpy.ndarray. dtype: [
            ('idx', 'i4'), ('open', 'f8'), ('close', 'f8'), ('high', 'f8'), ('low', 'f8'), ('volume', 'f8')
        ]
        I.e.: [
            (47729.99, 48495.  , 47716.9 , 32, 16145.34576)
            (45000.  , 45000.  , 44990.  ,  4,  3689.66919)
            (43996.5 , 44141.37, 43307.85, 25, 12609.72691)
            (43473.45, 43777.  , 43127.97, 20, 10403.64128)
            (42259.58, 42787.38, 42077.62, 21,  9709.83536)
            (41465.  , 42593.69, 41450.  , 28, 13145.51234)
        ]
        """

        # Combine and sort the sources
        src = np.concatenate((src_r, src_s))
        src = src[np.argsort(src["pivotPrice"])[::-1]]

        for curr_src in src:
            push = True
            idx_update = None

            if total_lines.size == 0:
                new_line = np.array(
                    [
                        (
                            curr_src["pivotPrice"],
                            curr_src["limitsUp"],
                            curr_src["limitsDown"],
                            curr_src["score"],
                            curr_src["accumulatedVolume"],
                        )
                    ],
                    dtype=total_lines.dtype,
                )
                total_lines = np.append(total_lines, new_line)
            else:
                for idx, curr_total_line in enumerate(total_lines):
                    dif = abs(
                        (
                            (curr_src["pivotPrice"] - curr_total_line["pivotPrice"])
                            / curr_src["pivotPrice"]  # noqa: W503
                        )
                        * 100  # noqa: W503
                    )

                    if dif <= threshold:  # update
                        push = False
                        idx_update = idx
                        break

                if push:  # push
                    new_line = np.array(
                        [
                            (
                                curr_src["pivotPrice"],
                                curr_src["limitsUp"],
                                curr_src["limitsDown"],
                                curr_src["score"],
                                curr_src["accumulatedVolume"],
                            )
                        ],
                        dtype=total_lines.dtype,
                    )
                    total_lines = np.append(total_lines, new_line)
                else:  # update
                    total_lines[idx_update]["pivotPrice"] = max(
                        curr_src["pivotPrice"], total_lines[idx_update]["pivotPrice"]
                    )
                    total_lines[idx_update]["limitsUp"] = max(
                        curr_src["limitsUp"], total_lines[idx_update]["limitsUp"]
                    )
                    total_lines[idx_update]["limitsDown"] = min(
                        curr_src["limitsDown"], total_lines[idx_update]["limitsDown"]
                    )
                    total_lines[idx_update]["score"] += curr_src["score"]
                    total_lines[idx_update]["accumulatedVolume"] += curr_src[
                        "accumulatedVolume"
                    ]

        return total_lines

    @classmethod
    def exec_pipeline(  # pylint: disable=dangerous-default-value
        cls,
        input_data: Dict[str, List[Union[float, int]]],
        cluster_threshold: float | None = 1.0,
        previous_sr: list | None = [],
    ) -> List[Dict[str, Union[int, float]]]:
        """
        Calculate support/resistance process pipeline.
        Return: list[dict[str, Union[int, float]]]. Support/Resistance calculated data. I.e.: [
            {'pivotPrice': 48184.0, 'limitsUp': 48495.0, 'limitsDown': 47347.53, 'score': 67, 'accumulatedVolume': 44687.15635},
            {'pivotPrice': 47534.34, 'limitsUp': 47874.98, 'limitsDown': 46763.68, 'score': 41, 'accumulatedVolume': 23864.294279999995},
            {'pivotPrice': 45000.0, 'limitsUp': 45000.0, 'limitsDown': 44700.0, 'score': 8, 'accumulatedVolume': 6010.54166},
            {'pivotPrice': 43996.5, 'limitsUp': 44141.37, 'limitsDown': 43100.0, 'score': 40, 'accumulatedVolume': 17909.42211},
            {'pivotPrice': 43473.45, 'limitsUp': 43580.0, 'limitsDown': 42697.01, 'score': 37, 'accumulatedVolume': 18501.36496},
            {'pivotPrice': 42819.86, 'limitsUp': 42850.0, 'limitsDown': 42041.7, 'score': 15, 'accumulatedVolume': 6355.429050000001},
            {'pivotPrice': 42259.58, 'limitsUp': 42787.38, 'limitsDown': 42017.33, 'score': 24, 'accumulatedVolume': 12506.830009999998},
            {'pivotPrice': 41619.99, 'limitsUp': 42365.48, 'limitsDown': 41115.0, 'score': 91, 'accumulatedVolume': 42944.990900000004},
            {'pivotPrice': 41071.29, 'limitsUp': 41157.26, 'limitsDown': 40753.88, 'score': 8, 'accumulatedVolume': 4474.11918}
        ]
        """

        data_np = np.array(list(input_data.values())).T

        # Previous lines (empty)
        dtype = [
            ("pivotPrice", "f8"),
            ("limitsUp", "f8"),
            ("limitsDown", "f8"),
            ("score", "i4"),
            ("accumulatedVolume", "f8"),
        ]
        previous_sr = np.array(previous_sr, dtype=dtype)

        max_point: np.ndarray[np.ndarray] = cls.__calc_points(
            data=data_np, extreme="max"
        )
        min_point: np.ndarray[np.ndarray] = cls.__calc_points(
            data=data_np, extreme="min"
        )
        clusters_r: np.ndarray[np.ndarray[np.ndarray]] = cls.__clustering(
            data=max_point, threshold=cluster_threshold, direction="resistance"
        )
        clusters_s: np.ndarray[np.ndarray[np.ndarray]] = cls.__clustering(
            data=min_point, threshold=cluster_threshold, direction="support"
        )
        recovery_line_r: np.ndarray = cls.__recovery_line(
            clusters=clusters_r, direction="resistance"
        )
        recovery_line_s: np.ndarray = cls.__recovery_line(
            clusters=clusters_s, direction="support"
        )
        s_r_calculated: np.ndarray = cls.__unify(
            src_r=recovery_line_r,
            src_s=recovery_line_s,
            threshold=cluster_threshold,
            total_lines=previous_sr,
        )

        s_r_calculated_list: list = [
            {
                "pivotPrice": row["pivotPrice"],
                "limitsUp": row["limitsUp"],
                "limitsDown": row["limitsDown"],
                "score": row["score"],
                "accumulatedVolume": row["accumulatedVolume"],
            }
            for row in s_r_calculated
        ]
        return s_r_calculated_list
